/*
	File: AfhVerInfo.c
	C source file for the AFH FilterMeister Version Info resource utilities.

    Copyright 1999 AFH Systems.

    FilterMeister is free software. You can redistribute and modify it
    under the terms of the MPL2 License.
    See: https://www.mozilla.org/en-US/MPL/2.0/
    Any changes to the FilterMeister files have to be published or sent
    to harry@thepluginsite.com. The filter plugin code itself does not
    need to be disclosed.

*/

#if !defined(BETA) && !defined(STANDALONE)

#include <windows.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stddef.h>     /* offsetof */

#include "AfhFM.h"

#define INIT_VERS_SIZE 2048
#define VERS_SIZE_INCREMENT 2048

#define VSHEADERSIZE (offsetof(VS_BLOCK_HEADER, szKey))

typedef struct scope_tag {
    struct scope_tag *pPrevScope;
    VS_BLOCK_HEADER *pCurBlock;
} SCOPE;

static VS_BLOCK_HEADER *pVS_VERSION_INFO = NULL;
static VS_BLOCK_HEADER *pVS = NULL;
static SCOPE *pScope = NULL;
static size_t cbBytesAllocated;


    ///// local helper functions /////

static int pad(int cb)
{
    return (cb + 3) & ~3;
} /*pad*/


static void pad32(void)
{
    //int cbVersSize = (char *)pVS - (char *)pVS_VERSION_INFO;
    //int padCount = -cbVersSize & 3;
	size_t cbVersSize = (char *)pVS_VERSION_INFO - (char *)pVS;
    size_t padCount = cbVersSize & 3;

    memset(pVS, 0, padCount);
    pVS = (VS_BLOCK_HEADER *)((char *)pVS + padCount);
} /*pad32*/


static void setBlockLength(void)
{
    //assert(pScope != NULL);
    pScope->pCurBlock->wLength = (WORD)( (char *)pVS - (char *)(pScope->pCurBlock) );
} /*setBlockLength*/


static void popScope(void)
{
    if (pScope != NULL) {
        SCOPE *p = pScope;
        pScope = pScope->pPrevScope;
        free(p);
    }
} /*popScope*/


static void cleanUp(void)
{
    //clean up after a fatal (allocation) error...
    if (pVS_VERSION_INFO != NULL) {
        free(pVS_VERSION_INFO);
        pVS_VERSION_INFO = NULL;
        pVS = NULL;
    }
    while (pScope != NULL) {
        popScope();
    }
} /*cleanUp*/


static BOOL pushScope(void)
{
    SCOPE *p = malloc(sizeof(SCOPE));
    if (p == NULL) {
        cleanUp();
        return FALSE;
    }
    p->pPrevScope = pScope;
    p->pCurBlock = pVS;
    pScope = p;
    return TRUE;
} /*pushScope*/


static BOOL makeRoomFor(int nBytes)
{
    size_t cbVersSize = (char *)pVS - (char *)pVS_VERSION_INFO;
    size_t cbSizeNeeded = cbVersSize + nBytes;
    if (cbSizeNeeded > cbBytesAllocated) {
        // Need more room!
        // Round up to a multiple of VERS_SIZE_INCREMENT...
        size_t cbSizeWanted = ((cbSizeNeeded + VERS_SIZE_INCREMENT - 1)/VERS_SIZE_INCREMENT) *
                           VERS_SIZE_INCREMENT;
        VS_BLOCK_HEADER *pNewVersHeader = realloc(pVS_VERSION_INFO, cbSizeWanted);
        if (pNewVersHeader == NULL) {
            cleanUp();
            return FALSE;
        }
        pVS_VERSION_INFO = pNewVersHeader;
        pVS = (VS_BLOCK_HEADER *)((char *)pVS_VERSION_INFO + cbVersSize);
        cbBytesAllocated = cbSizeWanted;
    }
    return TRUE;
} /*makeRoomFor*/


    ///// VS_init /////

BOOL VS_init(void)
{
    // Initialize data structures to create a new RT_VERSION resource in memory.
    // Returns TRUE if successful, else FALSE;
    pScope = NULL;
    pVS_VERSION_INFO = malloc(INIT_VERS_SIZE);
    if (pVS_VERSION_INFO == NULL) {
        cleanUp();
        return FALSE;
    }
    cbBytesAllocated = INIT_VERS_SIZE;
    pVS = pVS_VERSION_INFO;
    return TRUE;
} /*VS_init*/


    ///// VS_begin_binary_block /////

BOOL VS_begin_binary_block(const char *szKey, const void *pValue, int cbValueSize)
{
    if (!makeRoomFor(pad(VSHEADERSIZE + sizeof(wchar_t)*((int)strlen(szKey) + 1)) +
                     pad(cbValueSize))) {
        //memory allocation failed
        return FALSE;
    }
    if (!pushScope()) {
        //failed
        return FALSE;
    }
    pVS->wLength = 0;
    pVS->wValueLength = cbValueSize;
    pVS->wType = 0; //binary value
    mbstowcs(pVS->szKey, szKey, strlen(szKey)+1);
    pVS = (VS_BLOCK_HEADER *)&pVS->szKey[strlen(szKey)+1];
    setBlockLength();
    pad32();
    if (pValue != NULL) {
        memcpy(pVS, pValue, cbValueSize);
        pVS = (VS_BLOCK_HEADER *)((char *)pVS + cbValueSize);
        setBlockLength();
        pad32();
    }
    return TRUE;
} /*VS_begin_binary_block*/


    ///// VS_begin_string_block /////

BOOL VS_begin_string_block(const char *szKey, const char *szValue)
{
    int ccValueSize = szValue ? (int)strlen(szValue) + 1 : 0;

    if (!makeRoomFor(pad(VSHEADERSIZE + sizeof(wchar_t)*((int)strlen(szKey) + 1)) +
                     pad(sizeof(wchar_t)*ccValueSize))) {
        //memory allocation failed
        return FALSE;
    }
    if (!pushScope()) {
        //failed
        return FALSE;
    }
    pVS->wLength = 0;
    pVS->wValueLength = ccValueSize;
    pVS->wType = 1; //string (UNICODE) value
    mbstowcs(pVS->szKey, szKey, strlen(szKey)+1);
    pVS = (VS_BLOCK_HEADER *)&pVS->szKey[strlen(szKey)+1];
    setBlockLength();
    pad32();
    if (ccValueSize != 0) {
        mbstowcs((wchar_t *)pVS, szValue, ccValueSize);
        pVS = (VS_BLOCK_HEADER *)((wchar_t *)pVS + ccValueSize);
        setBlockLength();
        pad32();
    }
    return TRUE;
} /*VS_begin_string_block*/


    ///// VS_end_block /////

void VS_end_block(void)
{
    //assert(pScope !=NULL);
    popScope();
    if (pScope != NULL) {
        setBlockLength();
    }
} /*VS_end_block*/


    ///// VS_finish /////

VS_BLOCK_HEADER *VS_finish(void)
{
    //assert(pScope == NULL);
    return pVS_VERSION_INFO; //Caller must free this pointer after use!
} /*VS_finish*/


#endif //#if !defined(BETA) && !defined(STANDALONE)
