/*
 *  **MEMBER**=SLCLIBS:MD386VAXLIB
 *  **CMS**=RMX_MD86VAX
 *
 *  MDA_DISASSEMBLE disassembles x86 or (minimally) x64 code.
 *  One call disassembles one instruction (including any prefixes).
 *
 *   INBS	(char [])	contains the binary code to be disassembled (passed
 *						as a character array).
 *   INBL   (INT4)		length of INBS in bytes.
 *   RADIX  (INT4)		radix (10 or 16) to be used in most numeric output.
 *   OPTS   (UINT4)		option bits must be set as follows:
 *            bit 0:      1 if an attempt is to be made to display
 *                         operands symbolically (by calling
 *                         MDZ_EXPAND_SYMBOLIC_ADDRESS). NYI
 *            bit 1:      1 if CS & rIP are valid in the STATE structure.
 *            bit 2:      1 if all entries in STATE structure are valid.
 *            bit 3:      D-bit of the current code segment descriptor.
 *                         The default address size and operand size are
 *                         each 16 bits if this bit is 0, and 32 bits if
 *                         this bit is 1.
 *	          bit 4:      L-bit of the current code segment descriptor.
 *						   Enables 64-bit mode when set to 1, which also
 *                         overrides the effect of the D-bit (bit 3 above)
 *						   and sets the default address size to 64 bits
 *						   and the default operand size to 32 bits.
 *            bits 5-31:  Unused, should be 0.
 *   STATE  (struct *)	26-word structure containing the current contents of
 *						the CPU registers (see OPTS above).
 *   OUCS   (char [])	gets the ascii result (passed as a character array;
 *						the length indicated by the OUCL argument will
 *						not be exceeded).
 *   OUCL   (INT4)		(max) length of OUCS in bytes.
 *   NB     (INT2 *)	gets the number of bytes actually fetched from INBS,
 *						i.e., the true instruction length (should not exceeed
 *						15 for a valid x86/x64 instruction).
 *   NC     (INT2 *)	gets the number of bytes output into OUCS (not counting
 *						trailing blanks, though OUCS[NC] thru OUCS[OUCL-1] will 
 *						be blank-filled).
 *   IFLAGS (UINT4 *)	gets set as follows:
 *            bit 0:       1 if invalid instruction (UNK, FUNK, etc.).
 *            bit 1:       1 if CLI.
 *            bit 2:       1 if instruction has segment override prefix.
 *            bit 3:       1 if        ""       repeat prefix.
 *            bit 4:       1 if        ""       LOCK prefix.
 *            bit 5:       1 if MOV or POP to SS reg.
 *            bit 6:       1 if ESC (FPU instruction).
 *            bit 7:       1 if WAIT (or WAIT/ESC...).
 *            bit 8:       1 if HLT.
 *            bit 9:       1 if IN or OUT.
 *            bit 10:      1 if CALL.
 *            bit 11:      1 if RET.
 *            bit 12:      1 if INT 3 (opcode CC).
 *            bit 13:      1 if INT x (opcode CD).
 *            bit 14:      1 if INTO or BOUND (invalid in 64-bit mode).
 *            bit 15:      1 if IRET.
 *            bit 16:      1 if unconditional JMP.
 *            bit 17:      1 if conditional jump.
 *            bit 18:      1 if jump, call, or ret is FAR.
 *            bit 19:      1 if POPF.
 *            bit 20:      1 if instruction has (1 or more) operand size (66h) overrides.
 *            bit 21:      1 if instruction has (1 or more) address size (67h) overrides.
 *			  bit 22:      1 if redundant or misplaced prefixes. (Warning)
 *            bit 23       1 if misplaced REX prefix.
 *            bits 24-27:  all 0.
 *            bits 28-31:  offset of primary opcode (= # of prefix bytes),
 *                         or 15 if more than 14 prefix bytes.
 *
 *
 *  08Dec82 TEG  Written.
 *  24Oct88 ADG  Converted to 80386/80387.
 *  25Nov91 TEG  Tables rebuilt.
 *  04May97 AFH  Converted to C.
 *  14May97 AFH  1. IFLAGS were sometimes set incorrectly, based on
 *                  second byte of 2-byte opcode instead of first byte.
 *               2. Always display an explicit segment override, even if
 *                  it is the default.
 *               3. Set IFLAGS correctly for BOUND and ESC.
 *  18JUL98 AFH  1. Bits 20 and 21 in IFLAGS were not being set!
 *               2. Set "invalid instruction" flag in IFLAGS if MNERG or
 *                  MNESG starts with '?'.
 *               3. Return offset of primary opcode (i.e., # of prefix
 *                  bytes, including WAIT) in bits 28-31 of IFLAGS).
 *                  N.B. An offset of 15 indicates a dubious instruction!
 *  27Sep98 AFH  1. Fix bug which sometimes added a blank after a 16-bit
 *                  register name.
 *  17Jul12 AFH  1. Repeated Address-size and Operand-size prefixes are
 *					each cumulative (not toggling).  (Some manuals say
 *                  behavior with redundant prefixes is unspecified.)
 *  18Jul12 AFH  1. Add minimal support for x64 architecture.
 *               2. Set bit 22 in IFLAGS if redundant prefixes are detected.
 *               3. Set bit 23 in IFLAGS if a misplaced REX prefix.
 *  15Oct12 AFH	 1. Display x64 extended low-byte registers.
 *	31May14 AFH `1. Special handling for ARPL/MOVSXD in x64 mode.
 *
 *  TODO NEXT: display more x64 regs...
 ////////////////////////////////////////// show XMM regs for XMOVD/Q ops
 *
 *   --------------------------------------------------------------------
 *
 */

/************************* IMPLEMENTATION NOTES ************************/
/*
/*
/* Determining effective operand and address size
/* ==============================================
/*
/* For simplicity, we ignore the LMA bit and assume that the LMA bit
/* is set iff the CS.L bit is set, thus assuming that legacy and
/* long compatibility modes are essentially identical (as far as the
/* application programmer is concerned).  We thus distinguish only
/* 3 operating modes, which we herein call USE16, USE32, and USE64.
/* 
/* USE16 - includes legacy real mode, legacy virtual 8086 mode,
/*         legacy 16-bit protected mode, and long 16-bit
/*         compatibility mode.
/* USE32 - includes legacy 32-bit protected mode and long 32-bit
/*         compatibility mode.
/* USE64 - is long 64-bit mode.
/* 
/*                     +-----------------+--------------------------+
/*                     | eff. addr. size | effective operand size   |
/* +-----+-----+-------+--------+--------+--------+--------+--------+
/* |CS.L |CS.D |  mode |def.addr|  67H   |def.opnd| REX.W=1|  66H   |
/* +-----+-----+-------+--------+--------+--------+--------+--------+
/* |  0  |  0  | USE16 | 16-bit | 32-bit | 16-bit | ignored| 32-bit |
/* +-----+-----+-------+--------+--------+--------+--------+--------+
/* |  0  |  1  | USE32 | 32-bit | 16-bit | 32-bit | ignored| 16-bit |
/* +-----+-----+-------+--------+--------+--------+--------+--------+
/* |  1  |  0  | USE64 | 64-bit | 32-bit | 32-bit*| 64-bit | 16-bit&|
/* +-----+-----+-------+--------+--------+--------+--------+--------+
/* |  1  |  1  |reserv.| ------ | ------ | ------ | ------ | ------ |
/* +-----+-----+-------+--------+--------+--------+--------+--------+
/* 
/* *-some opcodes are auto-64-bit operand size;
/*   64-bits if opcode is auto-64-bit.
/* &-if REX.W=0
/* 
/**************************************************************************/

// Turn off CRT security warnings...
#define _CRT_SECURE_NO_WARNINGS 1


#if defined(VAXC) || defined(__DECC)
/* Following pragma packs data structure members on byte  */
/* boundaries while aligning the base of the structure on */
/* a longword (as required, e.g., for SLCNET buffers).    */
#pragma nomember_alignment LONGWORD
/* Following pragma kills some spurious error messages    */
/* caused by the preceding pragma.                        */
#pragma message disable (NONMULTALIGN)
#endif

#if  defined(WIN32) || defined(WIN64)
 /* Following pragma packs data structure members on byte boundaries. */
#pragma pack(push, 1)
#endif


#include <string.h>                     /* memcpy.                            */
#include "plmcvax.h"
#include "mda386tb.h"
#include "mda.h"


#if defined(VAXC) || defined(__DECC)

UINT4 lib$extzv(int pos, int nbits, UINT4 val);

#else
/*
 *  Define lib$extzv() as a macro for efficiency on non-DEC platforms.
 */
#define lib$extzv(pos,nbits,val) (((val) >> (pos)) & (((1 << (nbits)) - 1)))
#endif

/********/

#define min0(a,b) ((a) < (b) ? (a) : (b))

#define descr(sz) (sz),sizeof(sz)

/* adjust x86 byte regs(index 0) to x64 low-byte regs(index 7) if REX prefix is present */
#define REG8FIX(siz) ( (nrex && (siz)==0) ? 7 : (siz) )

static INT4 chtrunb(const char c[], INT4 len)
{
    int i;

    for (i = len; i > 0; --i)
    {
        if (c[i-1] != ' ') break;
    }
    return i;
}

static void strassn(char c1[], INT4 len1, const char c2[], INT4 len2)
{
    int i;

    for (i = 0; i < len1; ++i)
        c1[i] = ( i < len2 ? c2[i] : ' ');
}

/*
 *  MDISA86S.C contains character string fiddle subroutines
 *   MDFETBS, MDFETBZ, MDADDCH, MDADDCHT, MDADDCHTB,
 *   MDADDDEC, MDADDDECP, MDADDHEX, MDADDHEXP, MDADDDH, MDADDDHP
 *  which are used by the 8086/i286/i386/i486 disassembler MDA.C.
 *
 *  15Dec82 TEG  Written.
 *  07May97 AFH  Converted to C.
 *
 *    -------------------------------------------------------------------
 *
 *  MDFETBS(INBS, INBL, IN, NBYT, INOV)
 *  MDFETBZ(INBS, INBL, IN, NBYT, INOV)
 *
 *  Fetch at most 4 bytes out of character array;  return those bytes
 *  right-adjusted in a longword (I*4);  update pointer.
 *  MDFETBS sign-extends;  MDFETBZ zero-extends.
 *
 *   INBS      character array from which bytes are to be extracted;
 *              no access will occur past the end of INBS;  any bytes
 *              that would be past the end of INBS will be returned as 0.
 *   INBL      (I*4) length of INBS in bytes.
 *   IN        (I*4) character subscript of lowest-addressed byte to be
 *              extracted from INBS;  IN will be incremented by number of
 *              bytes fetched.
 *   NBYT      (I*4) number of bytes to be fetched  (0 .LE. NBYT .LE. 4;
 *              if NBYT=0 then result will be 0).
 *   INOV      (I*1) if an attempt was made to fetch bytes past the end
 *              of INBS, then a 1-bit will be or'ed into the low-order
 *              bit of INOV;  otherwise, INOV will remain unchanged.
 */
static INT4 mdfetbs(const char inbs[], INT4 inbl, INT4 *in, INT4 nbyt, INT1 *inov)
{
      INT4 n, val;

      n = inbl - *in + 1;
      if (n < nbyt) {
        *inov |= 1;
      } else {
        n = nbyt;
      }
      switch(n)
      {
      case 1:
          val = *(INT1 *)&inbs[*in-1];
          break;
      case 2:
          val = *(INT2 *)&inbs[*in-1];
          break;
      case 3:
          val = ((*(INT4 *)&inbs[*in-1] & 0xffffff) ^ 0x800000) - 0x800000;
          break;
      case 4:
          val = *(INT4 *)&inbs[*in-1];
          break;
      default:
          val = 0;
          n = 0;
      }
      *in += n;
      return val;
}

static UINT4 mdfetbz(const char inbs[], INT4 inbl, INT4 *in, INT4 nbyt, INT1 *inov)
{
      INT4 n;
      UINT4 val;

      n = inbl - *in + 1;
      if (n < nbyt) {
        *inov |= 1;
      } else {
        n = nbyt;
      }
      switch(n)
      {
      case 1:
          val = *(UINT1 *)&inbs[*in-1];
          break;
      case 2:
          val = *(UINT2 *)&inbs[*in-1];
          break;
      case 3:
          val = *(UINT4 *)&inbs[*in-1] & 0xffffff;
          break;
      case 4:
          val = *(INT4 *)&inbs[*in-1];
          break;
      default:
          val = 0;
          n = 0;
      }
      *in += n;
      return val;
}

/*
 *  Append substring to character variable;  update pointer.
 *
 *  MDADDCH appends substring as is;  MDADDCHT truncates right blanks
 *  from substring before appending it (if ADDC is all blanks, then
 *  MDADDCHT is a no-op);  MDADDCHTB is same as MDADDCHT plus appending
 *  one right blank.
 *
 *   OUTC      character array into which substring is to be stored;
 *              no storing will occur past the end of OUTC;  no bytes in
 *              OUTC past rightmost stored byte of substring will be
 *              altered.
 *   OUTL      (I*4) length of OUTC in bytes.
 *   OU        (I*4) character subscript of lowest-addressed byte in
 *              OUTC to be overwritten;  OU will be incremented by the
 *              number of bytes stored.
 *   ADDC      character array to be copied.
 *   ADDL      (I*4) length of ADDC in bytes.
 *
 *  15Dec82 TEG  Written.
 *
 */
static void mdaddch(char outc[], INT4 outl, INT4 *ou, const char addc[], INT4 addl)
{
      INT4 n;

      n = addl;
      n = min0(outl-*ou+1,n);
      if (n > 0) {
        memcpy(&outc[*ou-1], addc, n);
        *ou += n;
      }
      return;
}

static void mdaddcht(char outc[], INT4 outl, INT4 *ou, const char addc[], INT4 addl)
{
      INT4 n;

      n = chtrunb(addc,addl);
      n = min0(outl-*ou+1,n);
      if (n > 0) {
        memcpy(&outc[*ou-1], addc, n);
        *ou += n;
      }
      return;
}

/*
 *  Append substring to character variable
 *  (See MDADDCH/MDADDCHT).
 *
 *  15Dec82 TEG  Written.
 *
 */
static void mdaddchtb(char outc[], INT4 outl, INT4 *ou, const char addc[], INT4 addl)
{
      mdaddcht(outc,outl,ou,addc,addl);
      mdaddch(outc,outl,ou," ",1);
      return;
}

/*
 *  Convert 32-bit integer to ascii decimal (without blanks or leading
 *  zeroes, but preceded by minus sign if negative), and append resulting
 *  string to character variable;  update pointer.
 *  MDADDDECP is same as MDADDDEC, except that a nonnegative value will
 *  be preceded by a plus sign.
 *
 *   OUTC      character array into which the resulting string is to
 *              be stored;  no storing past end of OUTC will occur;  no
 *              bytes in OUTC past least significant digit of result 
 *              will be altered.
 *   OUTL      (I*4) length of OUTC in bytes.
 *   OU        (I*4) character subscript of lowest-addressed byte in 
 *              OUTC to be overwritten;  OU will be incremented by the
 *              number of bytes stored.
 *   ADN       (I*4) integer to be converted.
 *
 *  15Dec82 TEG  Written.
 *
 */
static const UINT4 pw10[10] = {1,10,100,1000,10000,100000,1000000,10000000,
                               100000000,1000000000};

static void mdadddec(char outc[], INT4 outl, INT4 *ou, INT4 adn)
{
      int i;
      UINT4 a;

      if (*ou <= outl)
      {
        if (adn < 0) {
          a = -adn;
          outc[(*ou)++-1] = '-';
        } else {
          a = adn;
        }
        for (i = 1; i < 10; ++i)
        {
          if (a < pw10[i]) break;
        }
        for (i = i-1; i >= 0; --i)
        {
          if (*ou > outl) break;
          outc[(*ou)++-1] = a/pw10[i] + '0';
          a %= pw10[i];
        }
      }
      return;
}

static void mdadddecp(char outc[], INT4 outl, INT4 *ou, INT4 adn)
{
      int i;
      UINT4 a;

      if (*ou <= outl)
      {
        if (adn < 0) {
          a = -adn;
          outc[(*ou)++-1] = '-';
        } else {
          a = adn;
          outc[(*ou)++-1] = '+';
        }
        for (i = 1; i < 10; ++i)
        {
          if (a < pw10[i]) break;
        }
        for (i = i-1; i >= 0; --i)
        {
          if (*ou > outl) break;
          outc[(*ou)++-1] = a/pw10[i] + '0';
          a %= pw10[i];
        }
      }
      return;
}

/*
 *  Convert 32-bit longword to ascii unsigned hexadecimal (without blanks
 *  and with at most one leading zero, and with a trailing 'H'), and
 *  append resulting string to character variable;  update pointer.
 *  MDADDHEXP is same as MDADDHEX, except that ADN is considered to be 
 *  signed, and will always be preceded by a plus sign or a minus sign.
 *
 *   OUTC      character array into which the resulting string is to
 *              be stored;  no storing past end of OUTC will occur;  no
 *              bytes in OUTC past the trailing 'H' will be altered.
 *   OUTL      (I*4) length of OUTC in bytes.
 *   OU        (I*4) character subscript of lowest-addressed byte in
 *              OUTC to be overwritten;  OU will be incremented by the
 *              number of bytes stored.
 *   ADN       (I*4) longword to be converted.
 *   NBYTH     (I*4) number of significant bytes in ADN (needed for
 *              eliminating unwanted sign-extension).
 *
 *  15Dec82 TEG  Written.
 *
 */
static const UINT4 xtzv[5] = {0,      0xff,    0xffff,  0xffffff,0xffffffff};
static const  INT4 xtsg[5] = {0,0xffffff80,0xffff8000,0xff800000,0x80000000};
static const char hx[16] = {'0','1','2','3','4','5','6','7',
                            '8','9','A','B','C','D','E','F'};

static void mdaddhex(char outc[], INT4 outl, INT4 *ou, INT4 adn, INT4 nbyth)
{
      UINT4 mval;
      int ib, msd, i, n;
      char tch[10] = "000000000H";

      mval = adn & xtzv[nbyth];
      ib = 0;
      msd = 8;
      for (i = 8; i >=1; --i)
      {
        tch[i] = hx[lib$extzv(ib,4,mval)];
        if (tch[i] != '0') msd = i;
        ib += 4;
      }
      if (tch[msd] > '9') --msd;
      n = min0(10-msd,outl-*ou+1);
      if (n > 0) {
        memcpy(&outc[*ou-1], &tch[msd], n);
        *ou += n;
      }
      return;
}

static void mdaddhexp(char outc[], INT4 outl, INT4 *ou, INT4 adn, INT4 nbyth)
{
      UINT4 mval;
      int ib, msd, i, n;
      char tch[10] = "000000000H";

      if (*ou <= outl)
      {
        if (adn < 0 && adn >= xtsg[nbyth]) {
          mval = -adn;
          outc[*ou-1] = '-';
        } else {
          mval = adn & xtzv[nbyth];
          outc[*ou-1] = '+';
        }
        ++*ou;
        ib = 0;
        msd = 8;
        for (i = 8; i >=1; --i)
        {
          tch[i] = hx[lib$extzv(ib,4,mval)];
          if (tch[i] != '0') msd = i;
          ib += 4;
        }
        if (tch[msd] > '9') --msd;
        n = min0(10-msd,outl-*ou+1);
        if (n > 0) {
          memcpy(&outc[*ou-1], &tch[msd], n);
          *ou += n;
        }
      }
      return;
}

/*
 *  Convert 32-bit longword to either ascii decimal or ascii hexadecimal,
 *  depending on the switch BPHEX (see also MDADDDEC, MDADDHEX).
 *
 *   OUTC      character array into which the resulting string is to
 *              be stored; no storing past the end of OUTC will occur.
 *   OUTL      (I*4) length of OUTC in bytes.
 *   OU        (I*4) character subscript of lowest-addressed byte in
 *              OUTC to be overwritten;  OU will be incremented by the
 *              number of bytes stored.
 *   ADN       (I*4) longword to be converted.
 *   BPHEX     (I*1) switch:  1 = hexadecimal,  0 = decimal.
 *   NBYTH     (I*4) number of significant bytes in ADN (used only if
 *              hex; needed for eliminating unwanted sign-extension).
 *
 *  16Dec82 TEG  Written.
 *
 */
static void mdadddh(char outc[], INT4 outl, INT4 *ou, INT4 adn, INT1 bphex, INT4 nbyth)
{
      if (bphex) {
        mdaddhex(outc,outl,ou,adn,nbyth);
      } else {
        mdadddec(outc,outl,ou,adn);
      }
      return;
}

/*
 *  Convert 32-bit longword to either signed ascii decimal or signed
 *  ascii hexadecimal, depending on the switch BPHEX.  The converted
 *  result will always be preceded by a plus sign or a minus sign (see
 *  also MDADDDECP, MDADDHEXP).
 *
 *   OUTC      character array into which the resulting string is to
 *              be stored; no storing past the end of OUTC will occur.
 *   OUTL      (I*4) length of OUTC in bytes.
 *   OU        (I*4) character subscript of lowest-addressed byte in
 *              OUTC to be overwritten;  OU will be incremented by the
 *              number of bytes stored.
 *   ADN       (I*4) longword to be converted.
 *   BPHEX     (I*1) switch:  1 = hexadecimal,  0 = decimal.
 *   NBYTH     (I*4) number of significant bytes in ADN (used only if
 *              hex; needed for eliminating unwanted sign-extension).
 *
 *  16Dec1982 TEG  Written.
 *
 */
static void mdadddhp(char outc[], INT4 outl, INT4 *ou, INT4 adn, INT1 bphex, INT4 nbyth)
{
      if (bphex) {
        mdaddhexp(outc,outl,ou,adn,nbyth);
      } else {
        mdadddecp(outc,outl,ou,adn);
      }
      return;
}

/********/


UINT4 Mda_disassemble(const char inbs[], INT4 inbl, INT4 radix, UINT4 opts,
                      struct cpustate_struc *state,
                      char oucs[], INT4 oucl, INT2 *nb, INT2 *nc, UINT4 *iflags)
{
      UINT4 mda_disassemble, opcase;
	  // Character buffers, expanded for x64.  Are they long enough?
   ///char cmodrm[60], cmodrm1[24], cstrop[24], creg[6];   //previous, for x86
      char cmodrm[80], cmodrm1[40], cstrop[24], creg[6];   //new, for x64
	  // Unused chars in above buffers cause unneeded additional cpu time,
	  // so keep dimensions to a reasonable (if not tight!) upper bound.
	  // By my computation (for x64):
	  //  max cmodrm used = ?? chars
	  //  max cmodrm1 used = ?? chars
	  //  max cstrop used = 17 chars
	  //  max creg used = 4 chars
      char clbrak[1] = "[", crbrak[1] = "]", cplus[1] = "+",
           cdoll[1] = "$", colon[1] = ":", comma[1] = ",",
           cptr[3] = "PTR", clock[4] = "LOCK", cnear[4] = "NEAR",
           cfar[3] = "FAR", ccl[2] = "CL", cone[1] = "1";
      INT4 in, in1, ou, op, op1, fop, flgl, mne, off, immed, immed1, nprnds,
           sego, modrm, m2, m3, reg, sib, sxr, sba, sss,
           mopsz, regsz, immsz, immsz1, isreg, offsz, repmne,
           strop, waitsvin, cx, cx1, cx2, opszx,
		   nrex, nsego, nopr, nadr, nlck, nrep, nwait;
      INT2 flgw;
	  //For bopr and badr 0=Use16, 1=Use32, 3=Use64
      INT1 bopr, badr, b64, bofnr,
           bhex, bhexi, inov, REX_byte;
	  INT1 rexr, rexx, rexb;	/* 8 if corresponding bit set in REX_byte, else 0 */

/*
 *   --------------------------------------------------
 *
 *
 *  Init various flags and char string subscripts.
 *
 */
      mda_disassemble = 1;
#if 1
	  radix = 10;	//force decimal
#endif
      if (radix != 16) {
        bhex = 0;
      } else {
        bhex = 1;
      }
      *iflags = 0;
      in = 1;
      inov = 0;
      strassn(oucs,oucl," ",1);
      ou = 1;
	  b64 = lib$extzv(4,1,opts); //Use64 (L) bit

      nwait = 0;

	  do {	/* Pick up next (prefixed) opcode */
	    /* Reset prefixes to default */
		nsego = 0;
        nlck = 0;
        nrep = 0;
		nopr = 0;
		nadr = 0;
		nrex = 0;
		REX_byte = 0;
        do {
		  /*
		   *  Get a byte to consider as the next instruction op code or prefix
		   *  thereof.
		   */
          in1 = in;                                  /* Save index of primary opcode */
          op = mdfetbz(inbs,inbl,&in,1,&inov);
          op1 = op;                                  /* Save primary opcode.   */
          flgl = DECOD[op];
          flgw = DECODW[lib$extzv(27,5,flgl)];		//DCD_DCWX field
          mne = OPMNE[op];
          switch (opcase = lib$extzv(13,3,flgw)) {
          case 1:
            sego = lib$extzv(24,3,flgl);             /* Segment override.      */ /*DCD_REGNT field*/
            nsego++;
			if (nrex) *iflags |= 0x800000;			 /* misplaced REX          */  // set IFLAGS later ???
            break;
          case 2:
			nopr++;							         /* Operand size toggle.   */
			if (nrex) *iflags |= 0x800000;			 /* misplaced REX          */
            break;
          case 3:
			nadr++;									 /* Address size toggle.   */
			if (nrex) *iflags |= 0x800000;			 /* misplaced REX          */
            break;
          case 4:
            nlck++;									 /* LOCK prefix.           */
 			if (nrex) *iflags |= 0x800000;			 /* misplaced REX          */
            break;
          case 5:
          case 6:
            repmne = mne;                            /* REP prefix.            */
            nrep++;
			if (nrex) *iflags |= 0x800000;			 /* misplaced REX          */
            break;
          case 7:
            op = mdfetbz(inbs,inbl,&in,1,&inov);     /* Two-byte opcode.       */
            flgl = DECODT[op];
            flgw = DECODW[lib$extzv(27,5,flgl)];	//DCD_DCWX field
            mne = OPMNET[op];
            break;
		  case 0:
          default:									 /* One-byte opcode or REX */
		    if (b64) {
			  /* For x64, one-byte INC or DEC becomes a REX prefix */
			  /* Check for REX prefix */
			  if ((op1 & 0xF0) == 0x40) {
				/* REX = 40h thru 4Fh */ 
				nrex++;	/* count REX prefixes */
				REX_byte = op1;						 /* Save latest REX byte   */
				opcase = 8;							 /* 8 = "phantom" REX case */
			  }
			  /* For x64, one-byte ARPL op becomes MOVSXD */
			  else if (op1 == 0x63) {
				flgl = 0x0001C2;	/* fictitious DECOD entry */
				flgw = 0x0000;		/* fictitious DECODW enrty */
				mne = MNE_MOVSXD;	/* index of "MOVSXD" in MNEOP[] */
			  }
		    }
            break;                                /* None of the above. */
          } /*switch*/
	    } while (opcase != 0 && opcase != 7);	/* loop until 1- or 2-byte opcode found */
		/*
		 *  Check for WAIT, which is an instruction in itself unless 
		 *  followed by ESC.
		 */
	    if ((flgw & DCW_WAIT) != 0) {
		  /* Found (another) WAIT instruction */
		  nwait++;
          if (nwait == 1) {
		    /* First WAIT -- discard prefixes and check next instruction */
            waitsvin = in;	/* save index of byte following this (first) WAIT */
          } else {
		    /* Second WAIT in a row! */
            in = waitsvin;	/* go back to first WAIT     */
		                    /* and process the first WAIT as */
							/* an instruction in itself...   */
				// but the flags and prefixes are set for the 2nd WAIT?
				// which shouldn't matter, but maybe we should reset them?
          }
        }
	  } while ((flgw & DCW_WAIT) && (nwait == 1));	//// IS THIS RIGHT ????? C.F. 1.09f!!!

	  /*
	   *  Determine net effects of any prefixes.
	   */
	  if (b64) {
		// 64-bit mode
		if (REX_byte & REX_W) {	//or if auto 64-bit instruction and nopr==0???
		  bopr = 3; //Use64
		  // (nopr is ignored in presence of REX.W)
		}
		else {
		  bopr = 1; //Use32
		  if (nopr) bopr = 0;	//toggle bopr to Use16
		}
		badr = 3; //Use64
		if (nadr) badr = 1;		//toggle badr to Use32
	  }
	  else {
		// 16-bit or 32-bit mode
	    // (We assume the stack segment has the same D-bit setting as
	    // the data segment in 16/32-bit modes.)
		bopr = lib$extzv(3,1,opts);	//set bopr to default (Use16 or Use32)
		badr = bopr;				//set badr to default (same as bopr)
		if (nopr) bopr ^= 1;	    //toggle bopr between Use16 and Use32
		if (nadr) badr ^= 1;		//toggle badr between Use16 and Use32
      }

	  // Set REX flags.
	  //////rexw = (REX_byte >> 3) & 1;	//REX.W
	  rexr = (REX_byte << 1) & 8;	//REX.R
	  rexx = (REX_byte << 2) & 8;	//REX.X
	  rexb = (REX_byte << 3) & 8;	//REX.B

	  /*
	   *  Fetch and decode modrm/sib bytes if any.
	   */
      bofnr = 0;
      strassn(descr(cmodrm1)," ",1);
      cx1 = 1;
      if ((flgl & DCD_MODRM) != 0) {
        modrm = mdfetbz(inbs,inbl,&in,1,&inov);
        m2 = lib$extzv(6,2,modrm);
        m3 = lib$extzv(0,3,modrm);
        reg = lib$extzv(3,3,modrm);
        if (m2 != 3) {                          /* Memory, not register.      */
          if (badr) {                           /* Use32 or Use64             */
            if (m3 != 4) {                      /* No SIB.                    */
              if (m2 != 0 || m3 != 5) {
                isreg = ISRTV[1][m3];
                mdaddch(descr(cmodrm1),&cx1,descr(clbrak));
                mdaddcht(descr(cmodrm1),&cx1,descr(MNERG[1+badr][rexb+m3]));  //+4 max	//REX.B+m3
              } else {                          /* Naked offset.              */
                bofnr = 1;
                isreg = 3;
                m2 = 2;
              }
            } else {                            
			  /* Fetch & decode SIB.        */
              sib = mdfetbz(inbs,inbl,&in,1,&inov);
              sxr = rexx + lib$extzv(3,3,sib);		//+REX.X (fully decoded)
              sba = lib$extzv(0,3,sib);
              sss = lib$extzv(6,2,sib);
              if (m2 != 0 || sba != 5) {        
				/* Base + scaled index + offset. */
                isreg = ISRTV[1][sba];	//default seg reg	//depends on REX???
                mdaddch(descr(cmodrm1),&cx1,descr(clbrak));
                mdaddcht(descr(cmodrm1),&cx1,descr(MNERG[1+badr][rexb+sba])); //+4 max //REX.B+sba
                if (sxr != 4) { //sxr != rSP (but R12 is okay as index)
                  mdaddch(descr(cmodrm1),&cx1,descr(cplus));
                  mdaddcht(descr(cmodrm1),&cx1,descr(MNERG[1+badr][sxr])); //+4 max //REX.X|sxr
                  mdaddcht(descr(cmodrm1),&cx1,descr(MNESSS[sss]));
                }
              } else {
                isreg = 3;
                if (sxr != 4) { //sxr != rSP (but R12 is okay as index) 
				  /* Scaled index + offset.     */
                  mdaddch(descr(cmodrm1),&cx1,descr(clbrak));
                  mdaddcht(descr(cmodrm1),&cx1,descr(MNERG[1+badr][sxr])); //+4 max //REX.X|sxr
                  mdaddcht(descr(cmodrm1),&cx1,descr(MNESSS[sss]));
                } else {                        /* Naked offset.              */
                  bofnr = 1;
                }
                m2 = 2;
              }
            }
            offsz = m2;
            if (offsz == 2) offsz = 4;	//X64 sometimes offsz = 8 ????
            off = mdfetbs(inbs,inbl,&in,offsz,&inov);
            if (off != 0 && !bofnr) {	//leave the case 0 in, to show forced disp with rBP?
				// X64: handle case of disp64 ???
              mdadddhp(descr(cmodrm1),&cx1,off,bhex,offsz);
            }
          } else {                              /* USE16                      */ //badr=0 => b64=0 hence no REX!!!
            if (m2 != 0 || m3 != 6) {           /* Regs + offset.             */
              isreg = ISRTV[0][m3];
              mdaddch(descr(cmodrm1),&cx1,descr(clbrak));
              mdaddcht(descr(cmodrm1),&cx1,descr(MNEXR[m3]));
            } else {                            /* Naked offset.              */
              isreg = 3;
              bofnr = 1;
              m2 = 2;
            }
            offsz = m2;
            off = mdfetbs(inbs,inbl,&in,offsz,&inov);
            if (off != 0 && !bofnr) {
              mdadddhp(descr(cmodrm1),&cx1,off,bhex,offsz);
            }
          } /*USE16*/
          if (!bofnr) {
            mdaddch(descr(cmodrm1),&cx1,descr(crbrak));
          }
        } /*m2!=3*/
/*
 *  Get secondary opcode (non-ESC) if any.
 */
        if ((flgl & DCD_SCNDOP) != 0) {
          if ((DECODS[mne-1][reg] & DCD_SCNDOP) == 0) {
            flgl = (flgl & 0xffffffff-DCD_DCWX) |
                   (DECODS[mne-1][reg] & DCD_DCWX+DCD_PFHEX);
          } else {
            flgl = (flgl & DCD_PFHEX+DCD_SCNDOP+DCD_MODRM) |
                   DECODS[mne-1][reg];
          }
          flgw = DECODW[lib$extzv(27,5,flgl)];	//DCD_DCWX field
          mne = OPMNES[mne-1][reg];
        } /*DCD_SCNDOP*/
/*
 *  Get non-modrm offset if any.
 */
      } else if ((flgl & DCD_OFFNOMOD) != 0) {
        isreg = 3;
        bofnr = 1;                           /* Naked offset.              */
        if (badr) {
          offsz = 4;	//Use32 or Use64 (maybe 8??)
        } else {
          offsz = 2;	//Use16
        }
        off = mdfetbs(inbs,inbl,&in,offsz,&inov);
      } /*DCD_OFFNOMOD*/
/*
 *  If ESC (80387), then decode same;  else check dangling WAIT.
 */
      if ((flgl & DCD_ESC) != 0) {
        fop = reg*8 | (op & 7);
        if (m2 != 3) {
          flgl = (flgl & 0xffff0000) | (DECODF[fop] & 0xffff) | DCD_MODRM2;
          mne = OPMNEF[fop];
        } else {
          flgl = (flgl & 0xffff0000) | (DECODFR[fop] & 0xffff) | DCD_MODRM2;
          mne = OPMNEFR[fop];
          if ((flgl & DCD_SCNDOP) != 0) {
            mne = OPMNEFS[mne-1][m3];
          }
        }
        mne = mne + MNEBASF;
      } else if (nwait) {
        op = 0x9b;
        op1 = op;                                  /* Save primary opcode.   */
        flgl = DECOD[op];
        flgw = DECODW[lib$extzv(27,5,flgl)];	//DCD_DCWX field
        mne = OPMNE[op];
        in = waitsvin;
        in1 = in - 1;                               /* Index of primary opcode */
      }
/*
 *  Determine operand sizes.
 */
      opszx = flgl & DCD_OPSIZX;
      if (opszx & DCD_00200000)
      { /* repair discontiguous field! */
        opszx += DCD_OPSIZX + 1 - (DCD_00200000 * 2);
      }
      if (bopr) {	//X64 ??? Use32 or Use64
        mopsz = OPSZV[lib$extzv(12,4,OPSIZ[opszx])];
        regsz = OPSZV[lib$extzv(16,4,OPSIZ[opszx])];
        immsz = lib$extzv(20,4,OPSIZ[opszx]);
		//Quick kludge for X64 -- probably not right in all cases!!!
		if (bopr == 3) {//or if auto64 such as CALL NEAR, RET NEAR, JMP SHORT/NEAR, Jcc, JrCXZ, LOOPcc,
						//ENTER, LEAVE, POP, and PUSH.
			//Use64
			if (mopsz == 3 && mne != MNE_MOVSXD) {
			  mopsz = 5; //Change dword to qword for memory operand (except MOVSXD!)
			}
			if (regsz == 3) regsz = 5;	//Change dword to qword for register operand
			//For MOV reg64,imm64 the immsz also increases to 5 (64-bits)
			//For MOV RAX,moffset64 and MOV moffset64,RAX the disp is also 64-bits.
			//N.B. Some instructions such as CPUID do NOT promote to 64 bits.
		}
      } else {		// Use16
        mopsz = OPSZV[lib$extzv(0,4,OPSIZ[opszx])];
        regsz = OPSZV[lib$extzv(4,4,OPSIZ[opszx])];
        immsz = lib$extzv(8,4,OPSIZ[opszx]);
      }
/*
 *  Get immediate data if any.
 */
      if (immsz > 0) {
        if ((flgw & DCW_ENTER) == 0) {
          if ((0xffff-flgw & DCW_JMPCALL+DCW_FAR) != 0) {
            if ((flgl & DCD_IBZX+DCD_IBMOD32) == 0) {
              immed = mdfetbs(inbs,inbl,&in,immsz,&inov);
            } else {
              immed = mdfetbz(inbs,inbl,&in,immsz,&inov);
              ///if ((flgl & DCD_IBMOD32) != 0) immed &= 0x1f;	//0x3f for X64, or just display the imm8 per se and let the cpu truncate however it wants???
            }
          } else {
            if (bopr) {		//Use32 or Use64
              immsz1 = 4;
            } else {		//Use16
              immsz1 = 2;
            }
            immed = mdfetbz(inbs,inbl,&in,immsz1,&inov);
            immed1 = mdfetbz(inbs,inbl,&in,2,&inov);
          }
        } else {	//(flgw & DCW_ENTER)
          immed = mdfetbz(inbs,inbl,&in,2,&inov);
          ///immed1 = mdfetbz(inbs,inbl,&in,1,&inov) & 0x1f;	//X64 ??? //don't truncate, let the dirt shine through, but maybe warn???
        }
      }
      nprnds = lib$extzv(5,2,flgl);		//DCD_NOPRNDS field
/*
 *  Check whether we overflowed the input byte string.
 */
      if (inov) {                         /* input bytestring overflow */
        flgl = 0;
        flgw = 0;
        mne = 0;                              /* call it UNK               */
        mda_disassemble = 0;
      }
/*
 *  Further format the modrm/sib info, if any.
 */
      strassn(descr(cmodrm)," ",1);
      cx = 1;
      if ((flgl & DCD_MODRM) != 0 || bofnr) {
        if (m2 != 3 || bofnr) {
#ifdef SUPPRESS_DEFAULT_SEG_OVERRIDE
          if (nsego > 0 && sego != isreg) {
#else
          if (nsego) {
#endif
            mdaddch(descr(cmodrm),&cx,descr(MNESG[sego]));
            if (sego > 5) *iflags |= 1;   /*invalid*/	//X64 ???
          }
          if ((nprnds == 1 || immsz > 0 || bofnr ||
               mopsz != regsz ||
               (flgl & DCD_ESC) != 0 ||
               (flgw & DCW_FAR+DCW_JMPCALL+DCW_SHIFT) != 0
              )
              && (flgw & DCW_LEA) == 0)
          {	//needs a PTR override...
            mdaddchtb(descr(cmodrm),&cx,descr(MNEPT[mopsz-1]));	//X64 ???
            if (!bofnr) {
              mdaddch(descr(cmodrm),&cx,descr(cptr));
            } else {
              mdaddchtb(descr(cmodrm),&cx,descr(cptr));
            }
          }
          if (!bofnr) {
            mdaddch(descr(cmodrm),&cx,cmodrm1,cx1-1);
          } else {
            mdadddh(descr(cmodrm),&cx,off,bhex,offsz);	//base offset ???
          }
        } else {  /* m2 == 3 */
          if ((flgl & DCD_ESC) == 0) {
            if (mopsz >= 1 && mopsz <= 5) {	//X64 ???
              mdaddcht(descr(cmodrm),&cx,descr(MNERG[REG8FIX(mopsz-1)][rexb+m3])); //REX.B+m3 //+4 max
              if (mopsz == 4) *iflags |= 1;   /*invalid*/
            }
          } else if ((flgl & DCD_SCNDOP) == 0) {
            mdaddch(descr(cmodrm),&cx,descr(MNEST[m3])); /*ST(m3)*/		//X64 ??
          }
        }
/*
 *  Format string operands if any.
 */
      } else if ((flgw & DCW_STRING) != 0) {
        strop = lib$extzv(24,3,flgl);	/*DCD_REGNT field*/
        if ((flgl & DCD_DESTREG) != 0) {
          isreg = 3;  /*default seg reg is DS*/
          if (nsego) isreg = sego;
          mdaddch(descr(cmodrm),&cx,descr(MNESG[isreg]));
          if (isreg > 5) *iflags |= 1;   /*invalid*/	//X64 ???
          mdaddchtb(descr(cmodrm),&cx,descr(MNEPT[mopsz-1]));	//X64 ???
          mdaddch(descr(cmodrm),&cx,descr(cptr));
          mdaddch(descr(cmodrm),&cx,descr(clbrak));
          mdaddcht(descr(cmodrm),&cx,descr(MNERG[1+badr][6])); /*rSI*/	//+3 max
          mdaddch(descr(cmodrm),&cx,descr(crbrak));
        } else {  //could combine the if-else parts!!
          mdaddch(descr(cmodrm),&cx,descr(MNESG[0]));	/*ES:*/
          mdaddchtb(descr(cmodrm),&cx,descr(MNEPT[mopsz-1]));	//X64 ???
          mdaddch(descr(cmodrm),&cx,descr(cptr));
          mdaddch(descr(cmodrm),&cx,descr(clbrak));
          mdaddcht(descr(cmodrm),&cx,descr(MNERG[1+badr][7])); /*rDI*/	//+3 max
          mdaddch(descr(cmodrm),&cx,descr(crbrak));
        }
        strassn(descr(cstrop)," ",1);										//0
        cx2 = 1;
        if (strop >= 3 && strop <= 4) {	//MOVS or CMPS
          if ((flgl & DCD_DESTREG) != 0) {
            mdaddch(descr(cstrop),&cx2,descr(MNESG[0])); /*ES:*/			//+3
            mdaddchtb(descr(cstrop),&cx2,descr(MNEPT[mopsz-1]));  //X64??	//+6 max
            mdaddch(descr(cstrop),&cx2,descr(cptr));						//+3
            mdaddch(descr(cstrop),&cx2,descr(clbrak));						//+1
            mdaddcht(descr(cstrop),&cx2,descr(MNERG[1+badr][7])); /*rDI*/	//+3 max
            mdaddch(descr(cstrop),&cx2,descr(crbrak));						//+1, max total=17
          } else { //could combine the if-else parts!!
            isreg = 3; /*default seg reg is DS*/
            if (nsego) isreg = sego;
            mdaddch(descr(cstrop),&cx2,descr(MNESG[isreg]));				//+3
            if (isreg > 5) *iflags |= 1;   /*invalid*/	//X64 ???
            mdaddchtb(descr(cstrop),&cx2,descr(MNEPT[mopsz-1]));  //X64??	//+6 max
            mdaddch(descr(cstrop),&cx2,descr(cptr));						//+3
            mdaddch(descr(cstrop),&cx2,descr(clbrak));						//+1
            mdaddcht(descr(cstrop),&cx2,descr(MNERG[1+badr][6])); /*rSI*/	//+3 max
            mdaddch(descr(cstrop),&cx2,descr(crbrak));						//+1, max total=17
          }
        }
      } else if ((flgw & DCW_XLAT) != 0) {
        isreg = 3;  /*default segreg is DS*/
        if (nsego) isreg = sego;
        mdaddch(descr(cmodrm),&cx,descr(MNESG[isreg]));
        if (isreg > 5) *iflags |= 1;   /*invalid*/	//X64 ???
        mdaddch(descr(cmodrm),&cx,descr(clbrak));
        mdaddcht(descr(cmodrm),&cx,descr(MNERG[1+badr][lib$extzv(24,3,flgl)]));	/*rBX*/ /*DCD_REGNT field*/ //+4 max
        mdaddch(descr(cmodrm),&cx,descr(crbrak));
      }
/*
 *  Format the register operand or operands.
 */
      if ((flgl & DCD_ESC) == 0) {
        if (regsz >= 1 && regsz <= 5) {	//X64 ???
          if ((flgl & DCD_MODRM) != 0) {
            if ((flgl & DCD_SCNDOP) == 0) {
              if ((flgl & DCD_REGSEG+DCD_REGSPCL) == 0) {
                strassn(descr(creg),descr(MNERG[REG8FIX(regsz-1)][rexr+reg])); //X64 REX.R+reg //4 max
                if (regsz == 4) *iflags |= 1;   /*invalid*/	//X64
              } else {
                if ((flgl & DCD_REGSPCL) == 0) {
                  strassn(descr(creg),MNESG[reg],2);							//2
                  if (reg > 5) *iflags |= 1;   /*invalid*/	//X64??
                } else {
                  strassn(descr(creg),descr(MNESPRG[lib$extzv(24,3,flgl)-1][rexr+reg]));	/*DCD_REGNT field*/ //REX.R+reg //4 MAX
                }
              }
            } else if ((flgl & DCD_IMPEAX) != 0) {
              strassn(descr(creg),descr(MNERG[regsz-1][0])); /*rAX*/ //X64 ???	//3 max
              if (regsz == 4) *iflags |= 1;   /*invalid*/	//X64 ??
            }
          } else {
            if ((flgl & DCD_OPREG) != 0) {
              if ((flgl & DCD_IMPEAX) == 0) {
                if ((flgl & DCD_REGSEG) == 0) {
                  strassn(descr(creg),descr(MNERG[REG8FIX(regsz-1)][rexb+lib$extzv(24,3,flgl)])); /*DCD_REGNT field*/ //REX.B+opreg ??? //4 max
                  if (regsz == 4) *iflags |= 1;   /*invalid*/	//X64 ??
                } else {
                  strassn(descr(creg),MNESG[lib$extzv(24,3,flgl)],2);			/*DCD_REGNT field*/ //2
                  if (lib$extzv(24,3,flgl) > 5) *iflags |= 1;   /*invalid*/		/*DCD_REGNT field*/	//X64 ??
                }
              } else {
                if ((flgw & DCW_XLAT) == 0) {
                  mdaddcht(descr(cmodrm),&cx,descr(MNERG[REG8FIX(mopsz-1)][rexb+lib$extzv(24,3,flgl)]));	/*DCD_REGNT field*/ //REX.B+opreg ??? //+4 max
                  if (mopsz == 4) *iflags |= 1;   /*invalid*/	//X64 ???
                }
                strassn(descr(creg),descr(MNERG[regsz-1][0])); /*rAX*/ //X64 ??? //3 max
                if (regsz == 4) *iflags |= 1;   /*invalid*/		//X64 ???
              }
            } else if ((flgl & DCD_IMPEAX) != 0) {
              strassn(descr(creg),descr(MNERG[regsz-1][0])); /*rAX*/ //X64 ??? //3 max
              if (regsz == 4) *iflags |= 1;   /*invalid*/	//X64 ???
            } else if ((flgw & DCW_STRING) != 0 && strop <= 2) {	//INS or OUTS
              strassn(descr(creg),MNERG[1][2],2);  /* port = DX */				//2
            }
          }
        }
      } else {
        if ((flgl & DCD_SCNDOP) == 0) {
          strassn(descr(creg),"ST",2);											//2
        } else if ((flgl & DCD_OPREG) != 0) {
          mdaddcht(descr(cmodrm),&cx,descr(MNERG[1][0]));     /* AX */	//+2	//X64 ??
        }
      }
/*
 *  Output the opcode mnemonic, inserting 'N' if 80387 instruction with
 *  no preceding WAIT.  Insert prefixes if any before opcode mnemonic.
 *  Insert 'NEAR' or 'FAR' after opcode mnemonic if jmp or call or ret.
 */
      if (nlck) {
        mdaddchtb(oucs,oucl,&ou,descr(clock));
      }
      if (nrep) {
        if ((0xffff-flgw & DCW_STRING+DCW_STRENE) != 0) {
          mdaddchtb(oucs,oucl,&ou,MNEOP[repmne],3); /*REP*/
        } else {
          mdaddchtb(oucs,oucl,&ou,descr(MNEOP[repmne])); /*REPE or REPNE*/
        }
      }
      if ((flgl & DCD_ESC) == 0 || nwait) {
        if ((flgl & DCD_MNEMD) == 0) {
          mdaddchtb(oucs,oucl,&ou,descr(MNEOP[mne]));
        } else {
          mdaddchtb(oucs,oucl,&ou,descr(MNEOP[mne+bopr]));	//X64 ??? NYI
        }
        if( ((flgw & DCW_JMPCALL) != 0) ^
            ((flgw & DCW_RET) != 0) ) {
          if ((flgw & DCW_FAR) == 0) {
            mdaddchtb(oucs,oucl,&ou,descr(cnear));
          } else {
            mdaddchtb(oucs,oucl,&ou,descr(cfar));
          }
        }
      } else {	// (flgl & DCD_ESC) && nwait==0)
	    /* Insert "N" after initial "F" to form no-wait form */
		/* of x87 mnemonic (e.g. FINIT becomes FNINIT).      */
        mdaddch(oucs,oucl,&ou,MNEOP[mne],1);
        mdaddch(oucs,oucl,&ou,"N",1);
        mdaddchtb(oucs,oucl,&ou,&MNEOP[mne][1],sizeof(MNEOP[mne])-1);
      }
/*
 *  Output the rest of the instruction.  We divide what's left to do into
 *  gross cases according to number of operands.
 */
      bhexi = bhex;
#if 0
      if ((flgl & DCD_PFHEX) != 0) bhexi = 1;	//always display immediate vals in hex for some opcodes
#endif
      if (nprnds == 1) {
        if ((flgl & DCD_MODRM) != 0 ||
            (flgw & DCW_STRING) != 0) {
          mdaddch(oucs,oucl,&ou,cmodrm,cx-1);
        } else if (immsz > 0) {
          if ((flgw & DCW_JMPCALL) == 0) {
            if ((flgw & DCW_AAMAAD) == 0 || immed != 10) {
              mdadddh(oucs,oucl,&ou,immed,bhexi,immsz);
            }
          } else {
            if ((flgw & DCW_FAR) == 0) {
              mdaddch(oucs,oucl,&ou,descr(cdoll));
              immed1 = immed + in - 1;
              if (immed1 != 0) mdadddhp(oucs,oucl,&ou,immed1,bhexi,immsz);
            } else {
              mdadddh(oucs,oucl,&ou,immed1,bhexi,2);
              mdaddch(oucs,oucl,&ou,descr(colon));
              mdadddh(oucs,oucl,&ou,immed,bhexi,immsz1);
            }
          }
        } else {
          mdaddcht(oucs,oucl,&ou,descr(creg));
        }
      } else if (nprnds == 2) {
        if ((flgl & DCD_MODRM) != 0 || bofnr) {
          if ((flgl & DCD_DESTREG) != 0) {
            mdaddcht(oucs,oucl,&ou,descr(creg));
            mdaddch(oucs,oucl,&ou,descr(comma));
            mdaddch(oucs,oucl,&ou,cmodrm,cx-1);
          } else {
            mdaddch(oucs,oucl,&ou,cmodrm,cx-1);
            mdaddch(oucs,oucl,&ou,descr(comma));
            if (immsz > 0) {
              mdadddh(oucs,oucl,&ou,immed,bhexi,immsz);
            } else {
              if ((flgw & DCW_SHIFT) == 0) {
                mdaddcht(oucs,oucl,&ou,descr(creg));
              } else {
                if ((flgl & DCD_SHIFTCL) != 0) {
                  mdaddch(oucs,oucl,&ou,descr(ccl));
                } else if ((flgl & DCD_SHIFT1) != 0) {
                  mdaddch(oucs,oucl,&ou,descr(cone));
                }
              }
            }
          }
        } else {
          if (immsz > 0) {
            if ((flgw & DCW_ENTER) == 0) {
              if ((flgl & DCD_DESTREG) != 0) {
                mdaddcht(oucs,oucl,&ou,descr(creg));
                mdaddch(oucs,oucl,&ou,descr(comma));
                mdadddh(oucs,oucl,&ou,immed,bhexi,immsz);
              } else {
                mdadddh(oucs,oucl,&ou,immed,bhexi,immsz);
                mdaddch(oucs,oucl,&ou,descr(comma));
                mdaddcht(oucs,oucl,&ou,descr(creg));
              }
            } else {
              mdadddh(oucs,oucl,&ou,immed,bhexi,2);
              mdaddch(oucs,oucl,&ou,descr(comma));
              mdadddh(oucs,oucl,&ou,immed1,bhexi,1);
            }
          } else if ((flgw & DCW_STRING) == 0) {
            if ((flgl & DCD_DESTREG) != 0) {
              mdaddcht(oucs,oucl,&ou,descr(creg));
              mdaddch(oucs,oucl,&ou,descr(comma));
              mdaddch(oucs,oucl,&ou,cmodrm,cx-1);
            } else {
              mdaddch(oucs,oucl,&ou,cmodrm,cx-1);
              mdaddch(oucs,oucl,&ou,descr(comma));
              mdaddcht(oucs,oucl,&ou,descr(creg));
            }
          } else { //(flgw & DCW_STRING)
            if (strop >= 3) {	//not INS or OUTS
              mdaddch(oucs,oucl,&ou,cstrop,cx2-1);
              mdaddch(oucs,oucl,&ou,descr(comma));
              mdaddch(oucs,oucl,&ou,cmodrm,cx-1);
            } else {	//INS or OUTS
              if ((flgl & DCD_DESTREG) != 0) {
                mdaddcht(oucs,oucl,&ou,descr(creg));
                mdaddch(oucs,oucl,&ou,descr(comma));
                mdaddch(oucs,oucl,&ou,cmodrm,cx-1);
              } else {
                mdaddch(oucs,oucl,&ou,cmodrm,cx-1);
                mdaddch(oucs,oucl,&ou,descr(comma));
                mdaddcht(oucs,oucl,&ou,descr(creg));
              }
            }
          }
        }
      } else if (nprnds == 3) {
        if ((flgl & DCD_DESTREG) == 0) {
          mdaddch(oucs,oucl,&ou,cmodrm,cx-1);
          mdaddch(oucs,oucl,&ou,descr(comma));
          mdaddcht(oucs,oucl,&ou,descr(creg));
        } else {
          mdaddcht(oucs,oucl,&ou,descr(creg));
          mdaddch(oucs,oucl,&ou,descr(comma));
          mdaddch(oucs,oucl,&ou,cmodrm,cx-1);
        }
        mdaddch(oucs,oucl,&ou,descr(comma));
        if (immsz > 0) {
          mdadddh(oucs,oucl,&ou,immed,bhexi,immsz);
        } else if ((flgl & DCD_SHIFTCL) != 0) {
          mdaddch(oucs,oucl,&ou,descr(ccl));
        }
      }
/*
 *  Set bits in IFLAGS.
 */
      if (mne == 0 || mne == MNEBASF) {
        *iflags |= 1;                                    /* UNK/FUNK      */
      }
      if (nsego) *iflags |= 4;                           /* segov prefix  */
      if (nrep) *iflags |= 8;                            /* rep prefix    */
      if (nlck) *iflags |= 0x10;                         /* lock prefix   */
      if ((flgl & DCD_ESC) != 0) *iflags |= 0x40;        /* esc           */
      if (nwait) *iflags |= 0x80;                        /* wait          */
      if ((flgw & DCW_INOUT) != 0) {
        *iflags |= 0x200;                                /* in/out        */
      } else if ((flgw & DCW_JMPCALL) != 0) {
        if ((flgw & DCW_JMPCOND) != 0) {
          *iflags |= 0x20000;                            /* condit jmp    */
        } else if (MNEOP[mne][0] == 'C') {
          *iflags |= 0x400;                              /* call          */
        } else {
          *iflags |= 0x10000;                            /* uncondit jmp  */
        }
        if ((flgw & DCW_FAR) != 0) {
          *iflags |= 0x40000;                            /* far           */
        }
      } else if ((flgw & DCW_RET) != 0) {
        *iflags |= 0x800;                                /* ret           */
        if ((flgw & DCW_FAR) != 0) {
          *iflags |= 0x40000;                            /* far           */
        }
      } else if ((flgw & DCW_IRET) != 0) {
        *iflags |= 0x8000;                               /* iret          */
      } else {
        if (op1 == 0xfa) *iflags |= 2;                   /* cli           */
        if (op1 == 0xf4) *iflags |= 0x100;               /* hlt           */
        if (op1 == 0xcc) *iflags |= 0x1000;              /* int 3         */
        if (op1 == 0xcd) *iflags |= 0x2000;              /* int x         */
        if (op1 == 0xce || op1 == 0x62)
                         *iflags |= 0x4000;              /* into/bound    */
        if (op1 == 0x9d) *iflags |= 0x80000;             /* popf          */
        if (op1 == 0x17 || (op1 == 0x8e && reg == 2))
                         *iflags |= 0x20;                /* pop/mov to SS */
      }
      if (nopr) *iflags |= 0x100000;					 /* operand size  */
      if (nadr) *iflags |= 0x200000;					 /* address size  */
	  if (nrex > 1 || nsego > 1 || nopr > 1 || nadr > 1 || nlck > 1 || nrep > 1)
						 *iflags |= 0x400000;			 /* redundant prefixes */
      *iflags |= min0(in1 - 1, 15) << 28;                /* # of prefix bytes */
/*
 *  Store values in NB & NC, and return.
 */
      *nb = in - 1;
      *nc = chtrunb(oucs,ou-1);
      return mda_disassemble;
} /* Mda_disassemble */

#ifdef WIN32
 /* restore default packing */
#pragma pack(pop)
#endif
